<?php

/**
 * Copyright 2019 IQRF Tech s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\Repository\Models;

use Iqrf\Repository\Entities\OsDpa;
use Iqrf\Repository\Entities\OsDpaAttributtes;
use Iqrf\Repository\Exceptions\CorruptedData;
use Iqrf\Repository\Exceptions\ServiceUnavailable;
use Nette\Utils\Strings;

/**
 * IQRF OS and DPA versions manager
 */
final class OsAndDpaManager extends BaseManager {

	/**
	 * IQRF OS and DPA versions path
	 */
	private const PATH = 'osdpa';

	/**
	 * Returns IQRF OS and DPA attributes as query parameters
	 * @param OsDpaAttributtes $attributes IQRF OS and DPA attributes
	 * @return array<string, string> Query parameters
	 */
	private function getQueryAttributes(OsDpaAttributtes $attributes): array {
		$parameters = $attributes->toQueryParameters();
		$query = [];
		foreach ($parameters as $key => $value) {
			$key = Strings::capitalize($key);
			$query['os' . $key] = $value;
			$query['dpa' . $key] = $value;
		}
		return $query;
	}

	/**
	 * Lists all IQRF OS and DPA versions
	 * @param OsDpaAttributtes|null $attributes IQRF OS and DPA version attributes
	 * @return array<OsDpa> IQRF OS and DPA versions
	 * @throws CorruptedData
	 * @throws ServiceUnavailable
	 */
	public function list(?OsDpaAttributtes $attributes = null): array {
		$query = $attributes !== null ? $this->getQueryAttributes($attributes) : [];
		$response = $this->apiClient->sendApiRequest(self::PATH, $query);
		return $this->parseResponse($response);
	}

	/**
	 * Parses the JSON API response
	 * @param array<array<string, string>> $response JSON API response
	 * @return array<OsDpa> IQRF OS and DPA versions
	 */
	private function parseResponse(array $response): array {
		$array = [];
		foreach ($response as $item) {
			$array[] = OsDpa::fromArray($item);
		}
		return $array;
	}

	/**
	 * Returns filtered IQRF OS and DPA versions
	 * @param string|null $os IQRF OS build
	 * @param string|null $dpa DPA versions
	 * @param OsDpaAttributtes|null $attributes IQRF OS and DPA version attributes
	 * @return array<OsDpa> IQRF OS and DPA versions
	 * @throws CorruptedData
	 * @throws ServiceUnavailable
	 */
	public function get(?string $os = null, ?string $dpa = null, ?OsDpaAttributtes $attributes = null): array {
		$query = $attributes !== null ? $this->getQueryAttributes($attributes) : [];
		if ($os !== null) {
			$query['os'] = $os;
		}
		if ($dpa !== null) {
			$query['dpa'] = $dpa;
		}
		$response = $this->apiClient->sendApiRequest(self::PATH, $query);
		return $this->parseResponse($response);
	}

}
