<?php

/**
 * Copyright 2019 IQRF Tech s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\Repository\Models;

use Iqrf\Repository\Entities\File;
use Iqrf\Repository\Entities\Files;
use Iqrf\Repository\Exceptions\DataNotFound;
use Iqrf\Repository\Exceptions\ServiceUnavailable;
use Iqrf\Repository\Utils\FileClient;
use Nette\Utils\Json;
use Nette\Utils\JsonException;

/**
 * Files manager
 */
final class FilesManager {

	/**
	 * @var FileClient HTTP(S) file client
	 */
	private $httpClient;

	/**
	 * @var bool Use credentials
	 */
	private $useCredentials = false;

	/**
	 * Constructor
	 * @param FileClient $httpClient HTTP(S) file client
	 */
	public function __construct(FileClient $httpClient) {
		$this->httpClient = $httpClient;
	}

	/**
	 * Sets the base path
	 * @param string $path Base path
	 */
	public function setPath(string $path): void {
		$this->httpClient->setPath($path);
	}

	/**
	 * Stets the credentials usage
	 * @param bool $useCredentials Use credentials
	 */
	public function setUseCredentials(bool $useCredentials): void {
		$this->useCredentials = $useCredentials;
	}

	/**
	 * Download the file and returns its content
	 * @param string $fileName File name
	 * @return string|null File content
	 * @throws DataNotFound
	 * @throws ServiceUnavailable
	 */
	public function download(string $fileName): ?string {
		$file = $this->get($fileName);
		if (!$file instanceof File) {
			return null;
		}
		return $this->httpClient->getFile($file->getName(), $this->useCredentials);
	}

	/**
	 * Return the file entity
	 * @param string $fileName File name
	 * @return File|null File
	 * @throws DataNotFound
	 * @throws JsonException
	 * @throws ServiceUnavailable
	 */
	public function get(string $fileName): ?File {
		$files = $this->list();
		$name = ltrim($fileName, './');
		foreach ($files->getFiles() as $file) {
			if ($file->getName() === $name) {
				return $file;
			}
		}
		return null;
	}

	/**
	 * Lists files
	 * @return Files Files
	 * @throws DataNotFound
	 * @throws JsonException
	 * @throws ServiceUnavailable
	 */
	public function list(): Files {
		$json = $this->httpClient->getFile('/__files__.json', $this->useCredentials);
		$array = Json::decode($json, Json::FORCE_ARRAY);
		return Files::fromArray($array);
	}

}
